

import jas.engine.Sim;
import jas.engine.SimModel;
import jas.events.ISimEventListener;
import jas.events.SimGroupEvent;
import jas.io.ParameterFrame;
import jas.space.IntGrid;
import jas.space.ObjGrid;

import java.util.Collections;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.ListIterator;
import java.util.Set;

public class DistrictModel extends SimModel implements ISimEventListener {

	public static final int MODEL_TYPE_FIRMS_WORKERS = 0;
	public static final int MODEL_TYPE_ONLY_FIRMS = 1;
	public static final int MODEL_TYPE_ONLY_WORKERS = 2;
	
	public static final int EVENT_ADD_ORDER = 1;
	public static final int EVENT_SHUFFLE = 2;
	public static final int EVENT_DISTRIBUTE_HIRING = 3;
	public static final int EVENT_FLUSH_DEAD = 4;
	public static final int EVENT_EXPAND_POPULATION_RANDOM = 5;
	public static final int EVENT_EXPAND_POPULATION_PROXIMITY = 6;
	
	private static int firmCounter = 0;
	
	public ObjGrid firmSpace;
	public ObjGrid workerSpace;
	public IntGrid firmSpaceVisibility;
	public IntGrid workerSpaceVisibility;
	public Set[][] visibility;
	
	public LinkedList<Firm> firms;
	public LinkedList<Worker> workers;
	
	private static int orderCounter = 0;
	
	//Parameters
	public int modelType;
	public int worldXSize, worldYSize;
	public int firmsNumber, workersNumber;
	public int fixedOrdersLength;
	public int numberOfSkills;
	public int hiringsPerTick;
	public boolean permutateFirms, noWorkers, proximityCreation;
	public boolean verbose = true;
	
	public double probabilityNewFirm, probabilityNewWorker;
	
	public double firm_startStorageLevel, firm_minStorageLevel;
	public double firm_incrementWhenWorking, firm_consumptionWhenWorking;
	public double firm_consumptionPerTurn;
	
	public double worker_startSubsitenceLevel, worker_consumptionPerTurn;
	public double worker_incrementWhenWorking, worker_minSubsitenceLevel;
	
	
	public DistrictModel() {
		super();
	}

	@Override
	public void setParameters() {
		fixedOrdersLength = 0;
		numberOfSkills = 5;
		hiringsPerTick = 10;
		
		permutateFirms = false;
		noWorkers = true;
		proximityCreation = true;
		
		probabilityNewFirm = 1;
		probabilityNewWorker = 0.8;
		
		worldXSize = 120;
		worldYSize = 120;
		
		firmsNumber = 30;
		workersNumber = 50;
		
		firm_startStorageLevel = 2.0;
		firm_incrementWhenWorking = 1.0;
		firm_consumptionPerTurn = 0.5;
		
		ParameterFrame pf = new ParameterFrame(this, "Model parameters", path + "District.pForm.xml");
		addSimWindow(pf);

	}

	@Override
	public void buildModel() {
		buildObjects();
		buildSchedule();
	}
	
	private void buildObjects() {
		firms = new LinkedList<Firm>();
		workers = new LinkedList<Worker>();
		
		firmSpace = new ObjGrid(worldXSize, worldYSize);
		workerSpace = new ObjGrid(worldXSize, worldYSize);
		firmSpaceVisibility = new IntGrid(worldXSize, worldYSize);
		workerSpaceVisibility = new IntGrid(worldXSize, worldYSize);
		visibility = new HashSet[worldXSize][worldYSize];
		for (int x = 0; x < worldXSize; x++)
			for (int y = 0; y < worldYSize; y++)
				visibility[x][y] = new HashSet();
		
		for (int i = 0; i < firmsNumber; i++)
			firms.addFirst(new Firm(firmCounter++, firmSpace, visibility, numberOfSkills, 
					firm_startStorageLevel, firm_minStorageLevel, firm_incrementWhenWorking,
					firm_consumptionWhenWorking, firm_consumptionPerTurn));
		
		for (int i = 0; i < workersNumber; i++)
			workers.add(new Worker(workerSpace, visibility, 
					worker_startSubsitenceLevel, worker_consumptionPerTurn,
					worker_incrementWhenWorking, worker_minSubsitenceLevel));
	}

	public void flushDeads() {
		for (ListIterator it = firms.listIterator(); it.hasNext(); ) {
			Firm firm = (Firm) it.next();
			if (firm.isDead()) {
				it.remove();
				System.out.println("Die at " + Sim.getAbsoluteTime() + " " + firm + " = " + firm.getCurrentStorageLevel() + " - " + firm.getInactivity());
			}
		}
		
	
		for (ListIterator it = workers.listIterator(); it.hasNext(); )
			if ( ((Worker) it.next()).isDead() )
				it.remove();
		
	}
	
	private void buildSchedule() {
		SimGroupEvent s = eventList.scheduleGroup(0, 1);
		
		s.addEvent(this, EVENT_ADD_ORDER);
		s.addEvent(this, EVENT_SHUFFLE);
		s.addEvent(this, EVENT_DISTRIBUTE_HIRING);
		if (noWorkers)
			s.addCollectionEvent(firms, Firm.EVENT_HIRE_AND_WORK_F);
		else
			s.addCollectionEvent(firms, Firm.EVENT_HIRE_AND_WORK_P);
		
		if (modelType != MODEL_TYPE_ONLY_WORKERS)
			s.addCollectionEvent(firms, Firm.EVENT_UPDATE_CONSUMPTION);
		
		if (modelType != MODEL_TYPE_ONLY_FIRMS)
			s.addCollectionEvent(workers, getObjectClass("Worker"), "updateConsumption");
		
		if (modelType != MODEL_TYPE_ONLY_WORKERS)
			s.addCollectionEvent(firms, Firm.EVENT_DELIVER_ORDERS);
		
		s.addCollectionEvent(firms, Firm.EVENT_EXPAND);
		s.addEvent(this, EVENT_FLUSH_DEAD);
		
		if (proximityCreation)
			s.addEvent(this, EVENT_EXPAND_POPULATION_PROXIMITY);
		else
			s.addEvent(this, EVENT_EXPAND_POPULATION_RANDOM);
		
		s.addEvent(this, "printStatus");
	}
	 
	public void printStatus() {
		for (ListIterator it = firms.listIterator(); it.hasNext(); ) {
			Firm firm = (Firm) it.next();
			System.out.println(firm + " = " + firm.getCurrentStorageLevel() + " Inact: " + firm.getInactivity());
		}
	}
	
	public void addOrderAnAllocate() {
		
		 Order order;
		 if (fixedOrdersLength != 0)
			 order = new Order(orderCounter++, numberOfSkills, fixedOrdersLength);
		 else
			 order = new Order(orderCounter++, numberOfSkills);
			 
		 int spec = order.getCurrentSpecification();

		 for (Firm firm : firms)
			if (firm.getSkill() == spec) {
				order.setAssignedTo(firm);
				firm.receiveOrder(order);
				return;
			}
		 //Nel caso non sia stato assegnato a nessuno muore
	}
	 
	 public void orderCompleted(Order order) {
		 
	 }
	 
	 public void shuffleFirms() {
		 if (permutateFirms)
			 permutateLinkedList(firms);
		 else
			 rotateLinkedList(firms);		 
	 }
	 
	 public void shuffleWorkers() {
		 permutateLinkedList(workers);
	 }
	 
	 private void permutateLinkedList(LinkedList list)
	 {
		 Collections.shuffle(list);
	 }
	   
	 @SuppressWarnings("unchecked")
	 private void rotateLinkedList(LinkedList list)
	 {
		 if (list.size() > 0)
			 list.addLast(list.removeFirst());
	 }
	   
	 public void distributeHirings() {
		 if (firms.size() == 0)
			 return;
		 
		 for (int i = 0; i < hiringsPerTick; i++) {
			 int choice = Sim.getRnd().getIntFromTo(0, firms.size() - 1);
			 firms.get(choice).incrementWorkersToHire();
		 }
	 }

	public void performAction(int id) {
		switch (id)
		{
			case EVENT_ADD_ORDER:
				addOrderAnAllocate();
				break;
			case EVENT_SHUFFLE:
				shuffleFirms();
				shuffleWorkers();
				break;
			case EVENT_DISTRIBUTE_HIRING:
				distributeHirings();
				break;
			case EVENT_FLUSH_DEAD:
				flushDeads();
				break;
			case EVENT_EXPAND_POPULATION_PROXIMITY:
				expandPopulationProximity();
				break;
			case EVENT_EXPAND_POPULATION_RANDOM:
				expandPopulationRandom();
				break;
			default:
				throw new IllegalArgumentException("The action " + id + " is not supported.");		
		}
	}
	
	public void expandPopulationRandom() {
		for (int i=0; i < numberOfSkills; i++){
			if (Sim.getRnd().getIntFromTo(0, 10000) < probabilityNewWorker * 10000)
				workers.add(new Worker(workerSpace, visibility, 
						worker_startSubsitenceLevel, worker_consumptionPerTurn,
						worker_incrementWhenWorking, worker_minSubsitenceLevel));
			
			if (Sim.getRnd().getIntFromTo(0, 10000) < probabilityNewFirm * 10000)
				firms.add(new Firm(firmCounter++, firmSpace, visibility, numberOfSkills, 
						firm_startStorageLevel, firm_minStorageLevel, firm_incrementWhenWorking,
						firm_consumptionWhenWorking, firm_consumptionPerTurn));
		}
	}
	
	public void expandPopulationProximity() {
		for (int i = 0; i < numberOfSkills; i++){
			if (modelType != MODEL_TYPE_ONLY_FIRMS)
				if (Sim.getRnd().getIntFromTo(0, 10000) < probabilityNewWorker * 10000)
				{
					if (workers.size() <= 0)
						workers.add(new Worker(workerSpace, visibility, 
								worker_startSubsitenceLevel, worker_consumptionPerTurn,
								worker_incrementWhenWorking, worker_minSubsitenceLevel));
					else
					{
						Worker worker = workers.get(Sim.getRnd().getIntFromTo(0, workers.size() - 1));
						workers.add(new Worker(workerSpace, visibility, 
								worker_startSubsitenceLevel, worker_consumptionPerTurn,
								worker_incrementWhenWorking, worker_minSubsitenceLevel, worker));
					}
				}
			
			if (Sim.getRnd().getIntFromTo(0, 10000) < probabilityNewFirm * 10000)
				firms.addLast(new Firm(firmCounter++, firmSpace, visibility, numberOfSkills, 
						firm_startStorageLevel, firm_minStorageLevel, firm_incrementWhenWorking,
						firm_consumptionWhenWorking, firm_consumptionPerTurn));
		}
	}
}
