

import jas.engine.Sim;
import jas.events.ISimEventListener;
import jas.space.ObjGrid;

import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;

public class Firm extends DomainPlayer implements ISimEventListener{
	
	public static final int EVENT_DELIVER_ORDERS = 1;
	public static final int EVENT_EXPAND = 2;
	public static final int EVENT_HIRE_AND_WORK_F = 3;
	public static final int EVENT_HIRE_AND_WORK_P = 4;
	public static final int EVENT_UPDATE_CONSUMPTION = 5;
	
	public static final int MAX_INACTIVITY = 15;
	
	private DistrictModel model;
	
	private int id;
	private int skill;
	private LinkedList<Order> inOrders;
	private LinkedList<Order> outOrders;
	
	private int workersToHire = 0;
	private int inactivity = 0;
	private double currentStorageLevel;      
	private double minimumStorageLevel;       
	private double incrementWhenWorking;       
	private double consumptionWhenWorking;       
	private double consumptionPerTurn;
	
	public Firm(int id, ObjGrid space, Set[][] domain, int numberOfSkills, double startStorageLevel,
			double minimumStorageLevel, double incrementWhenWorking,
			double consumptionWhenWorking, double consumptionPerTurn) {
		
		super(space, domain);
		setAtRandom();
		
		this.id = id;
		skill = Sim.getRnd().getIntFromTo(0, numberOfSkills - 1);
		inOrders = new LinkedList<Order>();
		outOrders = new LinkedList<Order>();
		
		this.currentStorageLevel = startStorageLevel;
		this.minimumStorageLevel = minimumStorageLevel;
		this.incrementWhenWorking = incrementWhenWorking;
		this.consumptionWhenWorking = consumptionWhenWorking;
		this.consumptionPerTurn = consumptionPerTurn;
		
		setExtension(Sim.getRnd().getIntFromTo(0, 100));
		
		model = (DistrictModel) Sim.engine.getModelWithID("DistrictModel");
	}

	public int getSkill() {
		return skill;
	}

	public void setSkill(int skill) {
		this.skill = skill;
	}	

	public void receiveOrder(Order anOrder) {
		inOrders.addLast(anOrder);
		if (model.verbose)
			System.out.println(this + " received order " + anOrder);
	}

	public int getWorkersToHire() {
		return workersToHire;
	}

	public void setWorkersToHire(int workersToHire) {
		this.workersToHire = workersToHire;
	}
	
	public void incrementWorkersToHire() {
		this.workersToHire++;
	}
	
	public void hireAndWorkF() {
		if (inOrders.size() > 0) { // if firm has order
			// order: we are working on the first of the list
			Order order = inOrders.removeFirst();
			order.incrementStage();
			if (order.getCurrentStage() < order.getLenght()) {
				if (order.getCurrentSpecification() != skill)
					outOrders.addLast(order);
				else
					inOrders.addFirst(order);
			} else {
				if (model.verbose)
					System.out.println("\nOrder completed: " + order + "!\n");
			}
			inactivity = 0;
		} else
			inactivity++;

		work();
		workersToHire = 0;
	}
	
	public void work() {
		currentStorageLevel += incrementWhenWorking - consumptionWhenWorking;
	}
	
	public int getInactivity() {
		return inactivity;
	}

	public void hireAndWorkP() {
		if (inOrders.size() > 0) { // if firm has order
			// order: we are working on the first of the list
			Order order = inOrders.removeFirst();
			order.incrementStage();
			if (order.getCurrentStage() < order.getLenght()) {
				if (order.getCurrentSpecification() != skill)
					outOrders.addLast(order);
				else
					inOrders.addFirst(order);
			} else
				System.out.println("Order completed.");
			inactivity = 0;
		} else
			inactivity++;

		//Set workers = getPlayersInMyDomain(Worker.class);
		List workers = model.workers;
		Iterator it = workers.iterator();
		for (int i = 0; i < workersToHire && it.hasNext(); i++)
		{
			Worker worker = (Worker) it.next();
			work();
			worker.work();
		}
		workersToHire = 0;
	}
	
	public double updateConsumption()  // compute consumption and returns level
	{
		currentStorageLevel -= consumptionPerTurn;
		
		if (currentStorageLevel < minimumStorageLevel || inactivity > MAX_INACTIVITY)
			delete();
			
		return currentStorageLevel;
	}
	
	public void expandDomain() {
		if (Sim.getRnd().getIntFromTo(0, 1) == 0) 
			increaseExtension();
	}
	
	public void deliverOrders() {
		Set neighbors = getPlayersInMyDomain(Firm.class);
		while (outOrders.size() > 0)
		{
			final Order order = outOrders.removeFirst();
			boolean found = false;
			if (neighbors.size() > 0)
				for (Iterator iterator = neighbors.iterator(); iterator.hasNext() && !found; )
				{
					Firm firm = (Firm) iterator.next();
					if (firm.getSkill() == order.getCurrentSpecification())
					{
						if (model.verbose)
							System.out.println("Delivered order " + order + " to " + firm);
						firm.receiveOrder(order);
						order.setAssignedTo(firm);
						found = true;
					}
				}
			if (!found) {
				delete();
				if (model.verbose)
					System.out.println("Lost order " + order);
			}
		}
	}

	public double getCurrentStorageLevel() {
		return currentStorageLevel;
	}

	@Override
	public String toString() {
		return "Firm " + id + " (" + skill +") ";
	}

	public void performAction(int id) {
		switch (id)
		{
			case EVENT_DELIVER_ORDERS:
				deliverOrders();
				break;
			case EVENT_EXPAND:
				expandDomain();
				break;
			case EVENT_HIRE_AND_WORK_F:
				hireAndWorkF();
				break;
			case EVENT_HIRE_AND_WORK_P:
				hireAndWorkP();
				break;			
			case EVENT_UPDATE_CONSUMPTION:
				updateConsumption();
				break;
		}
	}
}
